#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <string.h>

#include "./ili9163/ili9163lcd.h"
#include "./ili9163/colors.h"
#include "./spi/spi.h"
#include "defines.h"
#include "song.h"
#include "main.h"

struct Food{
	int X;
	int Y;
	int type;
};

enum Menu {
	Main,
	Classic,
	Small,
	Advanced,
	About
};

volatile uint32_t intrs = NONE;
volatile int32_t curNote = REST;
int posX = START_SQ_X;
int posY = START_SQ_Y;
int is_alive = ALIVE;
int running = RUNNING;
int snake_direction = RIGHT;
int snake_size = DEFAULT;
int headX = START_HX;
int headY = START_HY;
int tailX = START_TX;
int tailY = START_TY;
int count = NONE;
int food_type = RESET;
int PA0_pressed = 0;
int PA1_pressed = 0;
int PA2_pressed = 0;
int **map;
long score = NONE;
long steps = NONE;
struct Food food[FOOD_SIZE];
Menu menu = Main;

/*____________________________________________________________________________*/
//Set LCD screen ILI9163 128x128 SPI.
DigitalPin	pinSCK(&PORTB, PB7), pinSDA(&PORTB, PB5), pinA0(&PORTD, PD2),
			pinRESET(&PORTD, PD3), pinCS(&PORTD, PD4);
ILI9163 	lcd(&pinA0, &pinRESET, &pinCS);

FUSES =
{
	.low = 0xDC,
	.high = 0x9F,
	.extended = 0xEF,
};

/*______________INTERUPTS_____________________________________________________*/
// TIMER0 overflow.
ISR(TIMER0_OVF_vect)
{
    if (curNote == REST)
        intrs = 0;
    else
    {
        intrs++;
        if (intrs >= curNote)
        {
            PORTD ^= (1 << PD5);
            intrs = 0;
        }
    }
}

//PORTA interrupts.
ISR(PCINT0_vect)
{
	if (!(PINA & (1 << PA0)) && menu == Main)
	{
		PA0_pressed = 1;
	}

	if (!(PINA & (1 << PA1)) && menu == Main)
	{
		PA1_pressed = 1;
	}


	if (!(PINA & (1 << PA2)) && menu == Main)
	{
		PA2_pressed = 1;
	}
}

/*______________BUTTONS_______________________________________________________*/
//Initialise buttons.
void init_buttons(void)
{
    //LEFT
    DDRA &= ~(1 << PA0);
    PORTA |= (1 << PA0);

    //UP
    DDRA &= ~(1 << PA1);
    PORTA |= (1 << PA1);

    //DOWN
    DDRA &= ~(1 << PA2);
    PORTA |= (1 << PA2);

    //RIGHT
    DDRA &= ~(1 << PA3);
    PORTA |= (1 << PA3);

    //Test LED
    DDRD |= (1 << PD7);
}

/*______________SONG__________________________________________________________*/
//Play song note.
void play(int32_t note, uint32_t len)
{
    int i;
    curNote = note;
    for (i = 0; i< len; i++)
        _delay_ms(SEMIQUAVER_TIME);
    curNote = REST;
    _delay_ms(BREATH_TIME);
}

/*______________LCD MENUS_____________________________________________________*/
//Set main screen lcd.
void set_lcd(void)
{
	lcd.drawClear(BLACK);
	lcd.drawString(textX(7, 1), textY(1, 1), 2, GREEN, BLACK, TITLE1);
	lcd.drawString(textX(6, 1), textY(3, 1), 2, GREEN, BLACK, TITLE2);
	lcd.drawRectFilled(textX(5, 1), textY(6, 1), textX(13, 1), textY(1, 1), WHITE);
	lcd.drawRectFilled(textX(5, 1), textY(8, 1), textX(13, 1), textY(1, 1), WHITE);
	lcd.drawRectFilled(textX(5, 1), textY(10, 1), textX(13, 1), textY(1, 1), WHITE);
	lcd.drawRectFilled(textX(5, 1), textY(12, 1), textX(13, 1), textY(1, 1), WHITE);
	lcd.drawString(textX(7, 1), textY(6, 1), 1, BLACK, WHITE, TYPE1);
	lcd.drawString(textX(7, 1), textY(8, 1), 1, BLACK, WHITE, TYPE2);
	lcd.drawString(textX(7, 1), textY(10, 1), 1, BLACK, WHITE, TYPE3);
	lcd.drawString(textX(7, 1), textY(12, 1), 1, BLACK, WHITE, TYPE4);
	lcd.drawRectFilled(textX(posX, 1), textY(posY, 1), textX(1, 1), textY(1, 1), WHITE);
}

//Set "About game" menu screen.
void about_menu(void)
{
	lcd.drawClear(BLACK);
	lcd.drawString(textX(6, 1), textY(1, 1), 1, LIGHTGREY, BLACK, A_LINE1);
	lcd.drawRectFilled(textX(2, 1), textY(4, 1), textX(1, 1), textY(1, 1), WHITE);
	lcd.drawString(textX(4, 1), textY(4, 1), 1, WHITE, BLACK, A_LINE2);
	lcd.drawRectFilled(textX(2, 1), textY(6, 1), textX(1, 1), textY(1, 1), BLUE);
	lcd.drawString(textX(4, 1), textY(6, 1), 1, BLUE, BLACK, A_LINE3);
	lcd.drawRectFilled(textX(2, 1), textY(8, 1), textX(1, 1), textY(1, 1), RED);
	lcd.drawString(textX(4, 1), textY(8, 1), 1, RED, BLACK, A_LINE4);
	lcd.drawRectFilled(textX(2, 1), textY(10, 1), textX(1, 1), textY(1, 1), YELLOW);
	lcd.drawString(textX(4, 1), textY(10, 1), 1, YELLOW, BLACK, A_LINE5);
	lcd.drawRectFilled(textX(2, 1), textY(12, 1), textX(1, 1), textY(1, 1), MAGENTA);
	lcd.drawString(textX(4, 1), textY(12, 1), 1, MAGENTA, BLACK, A_LINE6);
	lcd.drawRectFilled(textX(2, 1), textY(14, 1), textX(1, 1), textY(1, 1), GREEN);
	lcd.drawString(textX(4, 1), textY(14, 1), 1, GREEN, BLACK, A_LINE7);
}

//Display score screen.
void display_score(void)
{
	char str[10];
	sprintf(str, "%ld", score * 10);
	lcd.drawClear(BLACK);
	lcd.drawString(textX(4, 1), textY(4, 1), 2, GREEN, BLACK, SCORE);
	lcd.drawString(textX(8, 1), textY(8, 1), 2, GREEN, BLACK, str);

	//Go back to main screen.
	while (1) {
		if (!(PINA & (1 << PA0))) {
			break;
		}
	}
}

/*______________SELECT________________________________________________________*/
//Move square in the main menu screen to select a game mode.
void move_square(void)
{
	//Move down.
	if (PA2_pressed) {
		posY += 2;
		if (posY >= 13)
			posY = 12;
		PA2_pressed = 0;

	}

	//Move up.
	if (PA1_pressed) {
		posY -= 2;
		if (posY <= 6)
			posY = 6;
		PA1_pressed = 0;
	}
}

/*______________GAME__________________________________________________________*/
//Setting to be done before game starts.
void init_game(void)
{
	int i;

	//Prepare map.
	map = (int **)calloc(SIZE, sizeof(int *));
	for (i = 0; i < SIZE; i++) {
		map[i] = (int *)calloc(SIZE, sizeof(int));
	}

	//Prepare states for snake.
	is_alive = ALIVE;
	running = RUNNING;
	snake_direction = RIGHT;
	headX = START_HX;
	headY = START_HY;
	tailX = START_TX;
	tailY = START_TY;
	score = NONE;
	snake_size = DEFAULT;
	food_type = RESET;

	//Clear LCD and redraw map.
	lcd.drawClear(BLACK);
	lcd.drawRect(textX(0, 1), textY(0, 1),
				textX(SIZE, 1), textX(SIZE, 1), WHITE);

	//Draw starting position for snake.
	draw_snake_start();
}

//Draw snake in its initial position.
void draw_snake_start(void)
{
	lcd.drawRectFilled(textX(tailX, 1), textY(tailY, 1),
						textX(1, 1), textX(1, 1), WHITE);
	lcd.drawRectFilled(textX(headX, 1), textY(headY, 1),
						textX(1, 1), textX(1, 1), WHITE);
	//Place snake on map.
	map[tailX][tailY] = 1;
	map[headX][headY] = 1;
}

//Add food to a random point.
void add_food(int place, int type)
{
	//Delete the oldest item added in food array.
	lcd.drawRectFilled(textX(food[place].X, 1), textY(food[place].Y, 1),
						textX(1, 1), textX(1, 1), BLACK);

	//Get random position and add type.
	food[place].X = rand() % 20;
	if (food[place].X == 0) {
		food[place].X++;
	}
	food[place].Y = rand() % 20;
	if (food[place].Y == 0) {
		food[place].Y++;
	}
	food[place].type = type;

	//Draw food.
	lcd.drawRectFilled(textX(food[place].X, 1), textY(food[place].Y, 1),
						textX(1, 1), textX(1, 1), food_colors[type]);
}

//Check if score changes.
int check_for_score(void)
{
	int i;

	//If any food coords mathces head coords then return true and delete square.
	for (i = 0; i < FOOD_SIZE; i++) {
		if (headX == food[i].X && headY == food[i].Y) {
			lcd.drawRectFilled(textX(food[i].X, 1), textY(food[i].Y, 1),
								textX(1, 1), textX(1, 1), BLACK);
			food[i].X = -20;
			food[i].Y = -20;
			food_type = food[i].type;
			return 1;
		}
	}

	//Score hasn't change.
	return 0;
}

//Check for collision with itself.
void check_for_collision(void)
{
	if ((snake_direction == UP && map[headX][headY - 1]) ||
		(snake_direction == DOWN && map[headX][headY + 1]) ||
		(snake_direction == LEFT && map[headX - 1][headY]) ||
		(snake_direction == RIGHT && map[headX + 1][headY])) {
		is_alive = !is_alive;
	}
}

//Change snake direction.
void change_direction(void)
{
	if (!(PINA & (1 << PA0)) && snake_direction != RIGHT) {
		snake_direction = LEFT;
		return;
	}
	if (!(PINA & (1 << PA1)) && snake_direction != DOWN) {
		snake_direction = UP;
		return;
	}
	if (!(PINA & (1 << PA2)) && snake_direction != UP) {
		snake_direction = DOWN;
		return;
	}
	if (!(PINA & (1 << PA3)) && snake_direction != LEFT) {
		snake_direction = RIGHT;
		return;
	}
}

//Change game settings in advanced mode.
void change_game_settings(void)
{
	//Increase size.
	snake_size++;

	//Check for each type of food.
	switch(food_type) {
		case 0:
			score++;
			break;
		case 1:
			score += 5;
			break;
		case 2:
			score -= 5;
			break;
		case 3:
			while (snake_size > 2) {
				snake_size--;
				change_tail();
			}
			break;
		case 4:
			snake_size--;
			score++;
			change_tail();
			break;
		case 5:
			snake_size--;
			score += 5;
			change_tail();
			break;
		default:
			break;
	}

	//Reset food_type global value.
	food_type = RESET;
}

//Draw snake in its new position.
void draw_snake(void)
{
	int i;
	int j;

	//Traverse map and draw at (i, j) with white if its value is 1.
	for (i = 0; i < SIZE; i++) {
		for (j = 0; j < SIZE; j++) {
			if (map[i][j]) {
				lcd.drawRectFilled(textX(i, 1), textY(j, 1),
									textX(1, 1), textX(1, 1), WHITE);
			}
		}
	}
	_delay_ms(100);
}

//Change snake tail.
void change_tail(void)
{
	//Remove old tail.
	map[tailX][tailY] = 0;
	lcd.drawRectFilled(textX(tailX, 1), textY(tailY, 1),
						textX(1, 1), textX(1, 1), BLACK);

	//Change tail coords. There is already 1 in the map matrix.
	if (map[tailX + 1][tailY]) {
		tailX++;
	} else {
		if (map[tailX - 1][tailY]) {
			tailX--;
		} else {
			if (map[tailX][tailY + 1]) {
				tailY++;
			} else {
				if (map[tailX][tailY - 1]) {
					tailY--;
				}
			}
		}
	}
}

//Move snake.
void move(int type)
{
	int got_food = 0;
	int aux_type = 0;

	//Generate food at random positions. Allowed max 3 items.
	steps++;
	if (steps % 20 == 0) {
		if (type == ADVANCED) {
			aux_type = rand() % 6;
		}
		add_food(count, aux_type);
		count++;
		if (count == 3) {
			count = 0;
		}
	}

	//Change snake coords if direction was changed, else continue with the
	//same state.
	switch (snake_direction) {
		case UP:
			headY--;
			break;
		case DOWN:
			headY++;
			break;
		case LEFT:
			headX--;
			break;
		case RIGHT:
			headX++;
			break;
		default:
			break;
	}

	//Check for score increase.
	got_food = check_for_score();

	//Move head.
	map[headX][headY] = 1;

	//Check for game type to know how size changes and how score changes.
	switch(type) {
		case SMALL:
			if (got_food) {
				score++;
			}
			change_tail();
			break;
		case CLASSIC:
			if (!got_food) {
				change_tail();
			} else {
				score++;
				snake_size++;
			}
			check_for_collision();
			break;
		case ADVANCED:
			if (!got_food) {
				change_tail();
			} else {
				change_game_settings();
			}
			check_for_collision();
			break;
		default:
			break;
	}

	//Draw snake on lcd.
	draw_snake();
}

//Play game by selecting a type.
void play_game(int type)
{
	int i;

	//Draw a message for starting game.
	lcd.drawString(textX(4, 1), textY(2, 1), 1, RED, GREEN, PRESS_R);

	//While snake is still alive, play game.
	while(is_alive) {
		//Check for first movement (first button pressed on game).
		if (!running) {
			if (!(PINA & (1 << PA3))) {
				lcd.drawString(textX(4, 1), textY(2, 1),
								1, BLACK, BLACK, PRESS_R);
				running = !running;
				_delay_ms(50);
			}
		} else {
			//Else change direction, score and size based on game type.
			change_direction();
			switch(type) {
				case SMALL:
					move(SMALL);
					break;
				case CLASSIC:
					move(CLASSIC);
					break;
				case ADVANCED:
					move(ADVANCED);
					break;
				default:
					break;
			}

			//Check for collision with borders.
			if (headX == 0 || headX == SIZE || headY == 0 || headY == 15) {
				break;
			}
		}
	}

	//Game ended. Free map.
	for (i = 0; i < SIZE; i++) {
		free(map[i]);
	}
	free(map);

	//Display score screen.
	display_score();

	//End.
	return;
}

/*______________MAIN__________________________________________________________*/
int main(void)
{
	int k = 0;

	menu = Main;

	//Setup clock divider. Timer0 overflows on counting to 256.
    //8Mhz / 1 (CS0=1) = 8000000 increments/sec. Overflows every 256, so 31250
    //overflow interrupts/sec.
    TCCR0B |= (1 << CS00);

    //Enable overflow interrupts
    TIMSK0 |= (1 << TOIE0);

    //PD5 as output for buzzer
    DDRD |= (1 << PD5);

    TCNT0 = 0;
    intrs = 0;
    curNote = REST;

	//Clear interrupts and enable the.
	//cli();
    sei();
	PCICR |= (1 << PCIE0);
	PCMSK0 |= (1 << PCINT0) | (1 << PCINT1) | (1 << PCINT2) | (1 << PCINT3);

	//Init LCD and buttons.
	spi_init();
	init_buttons();
	lcd.init(ILI9163::ROT0);
	set_lcd();

	//Play game by starting with main menu screen.
	while(1) {
		//play_song();
		if (menu == Main) {
			play(song[k], len[k]);
			k++;
			if (k >= sizeof(song)/sizeof(int)) {
				k = 0;
			}
			//Cursor position.
			int aux = posY;
			move_square();
			//If cursor moved, draw it to its new position.
			if (aux != posY) {
				lcd.drawRectFilled(textX(posX, 1), textY(aux, 1),
									textX(1, 1), textY(1, 1), BLACK);
				_delay_ms(50);
				lcd.drawRectFilled(textX(posX, 1), textY(posY, 1),
									textX(1, 1), textY(1, 1), WHITE);
			}
			//Check if a select was made.
			if (PA0_pressed) {
				PA0_pressed = 0;
				switch(posY) {
					case 6:
						menu = Classic;
						init_game();
						break;
					case 8:
						menu = Small;
						init_game();
						break;
					case 10:
						menu = Advanced;
						init_game();
						break;
					case 12:
						menu = About;
						about_menu();
						break;
					default:
						menu = Main;
						break;
				}
			}
		} else {
			//Show "About" menu.
			if (menu == About) {
				if (!(PINA & (1 << PA3))) {
					set_lcd();
					menu = Main;
				}
			} else {
				//Start "Small" game type (Size does not increase and no
				//powerups).
				if (menu == Small) {
					play_game(SMALL);
					set_lcd();
					menu = Main;
				} else {
					//Start classic snake game.
					if (menu == Classic) {
						play_game(CLASSIC);
						set_lcd();
						menu = Main;
					} else {
						//Star advanced game type (Size increases, remains the
						//same or resets and score changes based on powerups).
						if (menu == Advanced) {
							play_game(ADVANCED);
							set_lcd();
							menu = Main;
						} else {
							//Else nothing changes.
							menu = Main;
						}
					}
				}
			}
		}
	}

	//End.
	return 0;
}
